import math
import os
import sys

import keras.backend as K
import numpy
import scipy.io as io
from keras.layers import Conv2D
from keras.layers import Input, Activation, AveragePooling2D, Concatenate, GlobalAveragePooling2D, Dense, \
    Conv2DTranspose, Add, Lambda, Convolution2DTranspose
from keras.models import Model
from keras.optimizers import adam

model_high_name = ""
model_low_name = ""
model_high = None
model_low = None
scale_factor = 3  # to supress the gpu memory error
patch_size = 74
padding = 48
highImage = None
lowImage = None

num_filters = 32
num_filters_low = 32
num_res_blocks = 10
res_block_scaling=None
num_res_blocks_low = 8
def loss(y_true, y_pred):
    loss = K.mean(K.square(y_true - y_pred))
    return loss

conv_counter = 0
dilation_array = [1,1,1,2,2,2,2,4,4,4,4,8,8,4,4,4,4,2,2,2,2,1,1]

def model_layer_high(Adam):
    input = Input(shape=(None, None, 1))
    global conv_counter
    conv_counter = 0
    conv1 = Conv2D(num_filters, 3, padding='same', activation="relu", kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(input)
    conv_counter = conv_counter + 1

    conv2 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv1)
    conv_counter = conv_counter + 1
    conv2_activation = Activation('relu')(conv2)
    conv3 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv2_activation)
    conv_counter = conv_counter + 1
    conv3_merged = Add()([conv1, conv3])

    conv4 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv3_merged)
    conv_counter = conv_counter + 1
    conv4_activation = Activation('relu')(conv4)
    conv5 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv4_activation)
    conv_counter = conv_counter + 1
    conv5_merged = Add()([conv3_merged, conv5])

    conv6 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv5_merged)
    conv_counter = conv_counter + 1
    conv6_activation = Activation('relu')(conv6)
    conv7 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv6_activation)
    conv_counter = conv_counter + 1
    conv7_merged = Add()([conv5_merged, conv7])

    conv8 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv7_merged)
    conv_counter = conv_counter + 1
    conv8_activation = Activation('relu')(conv8)
    conv9 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv8_activation)
    conv_counter = conv_counter + 1
    conv9_merged = Add()([conv7_merged, conv9])

    conv10 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv9_merged)
    conv_counter = conv_counter + 1
    conv10_activation = Activation('relu')(conv10)
    conv11 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv10_activation)
    conv_counter = conv_counter + 1
    conv11_merged = Add()([conv9_merged, conv11])

    conv12 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv11_merged)
    conv_counter = conv_counter + 1
    conv12_activation = Activation('relu')(conv12)
    conv13 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv12_activation)
    conv_counter = conv_counter + 1
    conv13_merged = Add()([conv11_merged, conv13])
    conv13_merged = Add()([conv13_merged, conv9_merged])

    conv14 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv13_merged)
    conv_counter = conv_counter + 1
    conv14_activation = Activation('relu')(conv14)
    conv15 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv14_activation)
    conv_counter = conv_counter + 1
    conv15_merged = Add()([conv13_merged, conv15])
    conv15_merged = Add()([conv15_merged, conv7_merged])

    conv16 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv15_merged)
    conv_counter = conv_counter + 1
    conv16_activation = Activation('relu')(conv16)
    conv17 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv16_activation)
    conv_counter = conv_counter + 1
    conv17_merged = Add()([conv15_merged, conv17])
    conv17_merged = Add()([conv17_merged, conv5_merged])

    conv18 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv17_merged)
    conv_counter = conv_counter + 1
    conv18_activation = Activation('relu')(conv18)
    conv19 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv18_activation)
    conv_counter = conv_counter + 1
    conv19_merged = Add()([conv17_merged, conv19])
    conv19_merged = Add()([conv19_merged, conv3_merged])

    conv20 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv19_merged)
    conv_counter = conv_counter + 1
    conv20_activation = Activation('relu')(conv20)
    conv21 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv20_activation)
    conv_counter = conv_counter + 1
    conv21_merged = Add()([conv19_merged, conv21])
    conv21_merged = Add()([conv21_merged, conv1])

    conv22 = Conv2D(num_filters, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform",
                    use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(
        conv21_merged)
    conv_counter = conv_counter + 1
    conv23 = Conv2D(1, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv22)
    model = Model(input, conv23, name="low")
    model.compile(optimizer=Adam, loss=loss, metrics=['mean_squared_error'])
    return model

def model_layer_low(Adam):
    input = Input(shape=(None, None, 1))
    global conv_counter
    conv_counter = 0
    conv1 = Conv2D(num_filters_low, 3, padding='same', activation="relu", kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(input)
    conv_counter = conv_counter + 1

    conv2 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv1)
    conv_counter = conv_counter + 1
    conv2_activation = Activation('relu')(conv2)
    conv3 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv2_activation)
    conv_counter = conv_counter + 1
    conv3_merged = Add()([conv1, conv3])

    conv4 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv3_merged)
    conv_counter = conv_counter + 1
    conv4_activation = Activation('relu')(conv4)
    conv5 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv4_activation)
    conv_counter = conv_counter + 1
    conv5_merged = Add()([conv3_merged, conv5])

    conv6 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv5_merged)
    conv_counter = conv_counter + 1
    conv6_activation = Activation('relu')(conv6)
    conv7 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv6_activation)
    conv_counter = conv_counter + 1
    conv7_merged = Add()([conv5_merged, conv7])

    conv8 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv7_merged)
    conv_counter = conv_counter + 1
    conv8_activation = Activation('relu')(conv8)
    conv9 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv8_activation)
    conv_counter = conv_counter + 1
    conv9_merged = Add()([conv7_merged, conv9])

    conv10 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv9_merged)
    conv_counter = conv_counter + 1
    conv10_activation = Activation('relu')(conv10)
    conv11 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv10_activation)
    conv_counter = conv_counter + 1
    conv11_merged = Add()([conv9_merged, conv11])

    conv12 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv11_merged)
    conv_counter = conv_counter + 1
    conv12_activation = Activation('relu')(conv12)
    conv13 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv12_activation)
    conv_counter = conv_counter + 1
    conv13_merged = Add()([conv11_merged, conv13])
    conv13_merged = Add()([conv13_merged, conv9_merged])

    conv14 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv13_merged)
    conv_counter = conv_counter + 1
    conv14_activation = Activation('relu')(conv14)
    conv15 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv14_activation)
    conv_counter = conv_counter + 1
    conv15_merged = Add()([conv13_merged, conv15])
    conv15_merged = Add()([conv15_merged, conv7_merged])

    conv16 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv15_merged)
    conv_counter = conv_counter + 1
    conv16_activation = Activation('relu')(conv16)
    conv17 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv16_activation)
    conv_counter = conv_counter + 1
    conv17_merged = Add()([conv15_merged, conv17])
    conv17_merged = Add()([conv17_merged, conv5_merged])

    conv18 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv17_merged)
    conv_counter = conv_counter + 1
    conv18_activation = Activation('relu')(conv18)
    conv19 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv18_activation)
    conv_counter = conv_counter + 1
    conv19_merged = Add()([conv17_merged, conv19])
    conv19_merged = Add()([conv19_merged, conv3_merged])

    conv20 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv19_merged)
    conv_counter = conv_counter + 1
    conv20_activation = Activation('relu')(conv20)
    conv21 = Conv2D(num_filters_low, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv20_activation)
    conv_counter = conv_counter + 1
    conv21_merged = Add()([conv19_merged, conv21])
    conv21_merged = Add()([conv21_merged, conv1])

    conv22 = Conv2D(num_filters_low, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform",
                    use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(
        conv21_merged)
    conv_counter = conv_counter + 1
    conv23 = Conv2D(1, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv22)
    model = Model(input, conv23, name="low")
    model.compile(optimizer=Adam, loss=loss, metrics=['mean_squared_error'])
    return model

def predictHighFrame(img):
    global model_high
    input_batch = numpy.zeros((1, img.shape[0], img.shape[1], 1))

    input_batch[0, :, :, 0] = img[:, :]
    predicted = model_high.predict(input_batch, batch_size=1, verbose=1)
    predicted[predicted[:] > 1] = 1
    predicted[predicted[:] < 0] = 0
    predicted = predicted.astype(numpy.float)

    batch_predicted = numpy.zeros((img.shape[0], img.shape[1]), dtype=numpy.double)
    batch_predicted[:, :] = predicted[0, :, :, 0]
    io.savemat('high.mat', {'high': batch_predicted})


def predictLowFrame(img):
    global model_low
    input_batch = numpy.zeros((1, img.shape[0], img.shape[1], 1))

    input_batch[0, :, :, 0] = img[:, :]
    predicted = model_low.predict(input_batch, batch_size=1)
    predicted[predicted[:] > 1] = 1
    predicted[predicted[:] < 0] = 0
    predicted = predicted.astype(numpy.float)

    batch_predicted = numpy.zeros((img.shape[0], img.shape[1]), dtype=numpy.double)
    batch_predicted[:, :] = predicted[0, :, :, 0]
    io.savemat('low.mat', {'low': batch_predicted})


def predictHighFramePatch(img, i, j, oriWidth, oriHeight):
    global model_high
    input_batch = numpy.zeros((1, img.shape[0], img.shape[1], 1))

    input_batch[0, :, :, 0] = img[:, :]
    predicted = model_high.predict(input_batch, batch_size=1, verbose=1)
    predicted[predicted[:] > 1] = 1
    predicted[predicted[:] < 0] = 0
    predicted = predicted.astype(numpy.float)

    batch_predicted = numpy.zeros((img.shape[0], img.shape[1]), dtype=numpy.double)
    batch_predicted[:, :] = predicted[0, :, :, 0]
    startX = 0
    startY = 0
    if i != -1:
        if i != 0:
            startX = padding
        if j != 0:
            startY = padding
        highImage[i:i + patch_size, j:j + patch_size] = batch_predicted[startX:patch_size + startX,
                                                        startY:patch_size + startY]
    else:
        height, width = batch_predicted.shape[:2]
        if height > width:
            highImage[0:height, oriWidth - width + padding:oriWidth] = batch_predicted[0:height, padding:width]
        else:
            highImage[oriHeight - height + padding:oriHeight, 0:width - padding] = batch_predicted[padding:height,
                                                                                   0:width - padding]


def predictLowFramePatch(img, i, j, oriWidth, oriHeight):
    global model_low
    input_batch = numpy.zeros((1, img.shape[0], img.shape[1], 1))

    input_batch[0, :, :, 0] = img[:, :]
    predicted = model_low.predict(input_batch, batch_size=1)
    predicted[predicted[:] > 1] = 1
    predicted[predicted[:] < 0] = 0
    predicted = predicted.astype(numpy.float)

    batch_predicted = numpy.zeros((img.shape[0], img.shape[1]), dtype=numpy.double)
    batch_predicted[:, :] = predicted[0, :, :, 0]
    startX = 0
    startY = 0
    if i != -1:
        if i != 0:
            startX = padding
        if j != 0:
            startY = padding
        lowImage[i:i + patch_size, j:j + patch_size] = batch_predicted[startX:patch_size + startX,
                                                       startY:patch_size + startY]
    else:
        height, width = batch_predicted.shape[:2]
        if height > width:
            lowImage[0:height, oriWidth - width + padding:oriWidth] = batch_predicted[0:height, padding:width]
        else:
            lowImage[oriHeight - height + padding:oriHeight, 0:width - padding] = batch_predicted[padding:height,
                                                                                  0:width - padding]

def dividePatches(img):
    height, width = img.shape[:2]
    newHeight = height
    newWidth = width
    global highImage
    global lowImage
    highImage = numpy.zeros((height, width), dtype=numpy.double)
    lowImage = numpy.zeros((height, width), dtype=numpy.double)

    for i in range(0, newHeight, patch_size):
        for j in range(0, newWidth, patch_size):
            startX = 0
            startY = 0
            if i - padding < 0:
                startX = 0
            else:
                startX = i - padding

            if j - padding < 0:
                startY = 0
            else:
                startY = j - padding

            predictFrame = img[startX:i + patch_size + padding, startY:j + patch_size + padding]

            predictHighFramePatch(predictFrame, i, j, width, height)
            predictLowFramePatch(predictFrame, i, j, width, height)


if __name__ == "__main__":
    if sys.argv.__len__() == 5:
        dual_iso = sys.argv[1]
        model_high_name = sys.argv[2]
        model_low_name = sys.argv[3]
        print(sys.argv[4])
        patch_size = int(sys.argv[4])
        img = io.loadmat(dual_iso)['img']

        initial_lr = 0.0001
        Adam = adam(lr=initial_lr)
        model_high = model_layer_high(Adam)
        model_high.load_weights(model_high_name)

        model_low = model_layer_low(Adam)
        model_low.load_weights(model_low_name)

        width, height = img.shape[:2]

        if width > 1200 and height > 1200:
            dividePatches(img)
            io.savemat('high.mat', {'high': highImage})
            io.savemat('low.mat', {'low': lowImage})
        else:
            predictHighFrame(img)
            predictLowFrame(img)
    else:
        print("There are missing arguments")
